unit PathFind;

interface

uses
  Windows, Classes;

type
  TPath = array of TPoint;  //       .

  TPathGroup = record
    Path:TPath;
    Size:Integer;
  end;

  TPathMapCell = record   //       :
    Distance : Integer;   // -      
    Direction : Integer;  // - ,      .
  end;



  TPathMap = array of array of TPathMapCell;  //  .

  {       ,  
       (X,Y)   Direction.
    (X,Y) -- ,     -1.
        .   
       .}
  TGetCostFunc = Function(X,Y,Direction : Integer) : Integer;

{        .
 ,          
      .}
procedure MakePathMap(
  StartX,StartY : Integer;
  GetCostFunc : TGetCostFunc);

procedure InitMap(MapWidth,MapHeight:Integer);

procedure FillPathMap(X1,Y1,X2,Y2 : Integer);

{     ,   (X,Y)
       .}
procedure FindPathOnMap(var Pt:TPathGroup; X,Y : Integer);

{     (StartX,StartY)   (StopX,StopY).}
procedure  FindPath(
  var pt:TPathGroup;
  StartX,StartY,StopX,StopY : Integer;
  GetCostFunc : TGetCostFunc);

implementation

uses u_game, u_const, zglHeader;

type
  TWaveCell = record      //    . :
    X,Y : Integer;        // -  ,    ;
    Cost : Integer;       // -      ;
    Direction : Integer;  // - ,    .
  end;

  TWave = class
  private
    FData : array of TWaveCell;
    FPos : Integer;                //    .
    FCount : Integer;              //     .
                                   //     
                                   //    . 
                                   //  .
    FMinCost : Integer;
    function GetItem : TWaveCell;
  public
    property Item : TWaveCell read GetItem;    //   .
    property MinCost : Integer read FMinCost;  //   Cost
                                               //    .
    constructor Create;
    destructor Destroy; override;
    procedure Add(NewX,NewY,NewCost,NewDirection : Integer); //  
                                                             //   .
    procedure Clear;  //  .      
                      //   FCount.   
                      //      .
    function Start : Boolean;  //   .  
                               //   .  false,
                               //   .
    function Next : Boolean;   //     .
                               //  false,    .
  end;

var
  GetCostFunc : TGetCostFunc;  //     
                               //  ,   
                               // .
  MapWidth : Integer;          //  .
  MapHeight : Integer;         //  .
  MainMap : TPathMap;

constructor TWave.Create;
begin
  Clear;  //   .
end;

destructor TWave.Destroy;
begin
  FData:=NIL;  //  ,  .
  inherited Destroy;
end;

function TWave.GetItem : TWaveCell;
begin
  Result:=FData[FPos];  //    .
end;

procedure TWave.Add(NewX,NewY,NewCost,NewDirection : Integer);
begin
  if FCount >= Length(FData)            //      ,
  then                                  // 
    SetLength(FData,Length(FData)+256);  //   .  256
                                        //   .
  with FData[FCount] do
    begin
    X:=NewX;
    Y:=NewY;
    Cost:=NewCost;
    Direction:=NewDirection;

  end;
  if NewCost < FMinCost  //   NewCost  .
  then
    FMinCost:=NewCost;
  Inc(FCount);           //    .
end;

procedure TWave.Clear;   //      .
begin
  FPos:=0;
  FCount:=0;
  FMinCost:=High(Integer);
end;

function TWave.Start : Boolean;
begin
  FPos:=0;               //      
  Result:=(FCount > 0);  // .  false,   .
end;

function TWave.Next : Boolean;
begin
  Inc(FPos);                //    . 
  Result:=(FPos < FCount);  // false,    .
end;

{    .   
   .    Direction   0..7,
       .  ,   
 -1 ( ),      .
  FoolProof:      
      .
         
    .}
function GetCost(X,Y,Direction : Integer) : Integer;
begin
  Direction:=(Direction AND 7);
  if (X < 0) OR (X >= MapWidth) OR (Y < 0) OR (Y >= MapHeight)
  then
    Result:=-1
  else
    Result:=GetCostFunc(X,Y,Direction);
end;

{ :      X.}
function DirToDX(Direction : Integer) : Integer;
begin
  case Direction of
    0,4  : Result:=0;
    1..3 : Result:=1;
  else
    Result:=-1;
  end;
end;

{ :      Y.}
function DirToDY(Direction : Integer) : Integer;
begin
  case Direction of
    2,6  : Result:=0;
    3..5 : Result:=1;
  else
    Result:=-1;
  end;
end;

procedure InitMap(MapWidth,MapHeight:Integer);
begin
  PathFind.MapWidth:=MapWidth;
  PathFind.MapHeight:=MapHeight;
  SetLength(MainMap,MapHeight,MapWidth);
end;

{   --  .   ,
      .
         .}

procedure FillPathMap(X1,Y1,X2,Y2 : Integer);
var
  OldWave, NewWave : TWave;
  Finished : Boolean;
  n, x_, Dr:integer;
  I : TWaveCell;

  procedure PreparePathMap;  //      
  var                        //      Distance
    X,Y : Integer;           // (   )  -1,
  begin                      // ,      .
    for Y:=0 to (MapHeight-1) do
      for X:=0 to (MapWidth-1) do
        MainMap[Y,X].Distance:=-1;
  end;

  procedure TestNeighbours;  //  ,   ,
  var                        //      .
    X,Y,C,D : Integer;
  begin
    for D:=0 to 7 do
      begin
      X:=OldWave.Item.X+DirToDX(D);
      Y:=OldWave.Item.Y+DirToDY(D);
      C:=GetCost(X,Y,D);
      if (C >= 0) AND (MainMap[Y,X].Distance < 0) //      
      then                                       // , 
        NewWave.Add(X,Y,C,D);                    //   .
    end;
  end;

  procedure ExchangeWaves; //      
  var
    W : TWave;
  begin
    W:=OldWave;
    OldWave:=NewWave;
    NewWave:=W;
    NewWave.Clear;
  end;

begin
  PreparePathMap;             //      
                              //  .
  OldWave:=TWave.Create;
  NewWave:=TWave.Create;
  MainMap[Y1,X1].Distance:=0;  // ,    .
  OldWave.Add(X1,Y1,0,0);     //      
  TestNeighbours;             //      OldWave 
                              //  Cost = 0,  TestNeighbours  
                              // NewWave      
                              //    .
  Finished:=((X1 = X2) AND (Y1 = Y2));  //     
                                        //  .
  while NOT Finished do        // ,     .
    begin
    ExchangeWaves;             //      OldWave 
                               //  NewWave   
                               // .
    if NOT OldWave.Start then Break;  //  OldWave  . 
                                      //  ,   . .
      repeat                          //   .
        I:=OldWave.Item;                 //    .
        I.Cost:=I.Cost-OldWave.MinCost;  //   Cost  
                                         //   OldWave  ,
                                        //     (Cost=0).


        if I.Cost > 0                             // ,   ,
        then begin                                     //   NewWave 
                                                       //  Cost.

          NewWave.Add(I.X,I.Y,I.Cost,I.Direction);


        end else
          begin
          //     ,  .
          if MainMap[I.Y,I.X].Distance >= 0 then Continue;
          //     ,    Distance  
          //  .      
          //  +1.
          MainMap[I.Y,I.X].Distance:=MainMap[I.Y-DirToDY(I.Direction),I.X-DirToDX(I.Direction)].Distance+1;
          //  ,     .
          MainMap[I.Y,I.X].Direction:=I.Direction;
          //     .

          Finished:=((I.X = X2) AND (I.Y = Y2));
          if Finished then Break;
          //         
          // .   NewWave.
          TestNeighbours;
          end;
      until NOT OldWave.Next;  //  ,   
    end;                       //    OldWave;
  NewWave.Free;
  OldWave.Free;
end;

procedure MakePathMap(
  StartX,StartY : Integer;
  GetCostFunc : TGetCostFunc);
begin
  PathFind.GetCostFunc:=GetCostFunc;
  //   ,     ,  
  //    ,      .
  // ,       .
  FillPathMap(StartX,StartY,-1,-1);
end;

procedure FindPathOnMap(var Pt:TPathGroup; X,Y : Integer);
var
  Direction : Integer;
begin
  //Pt.Path:=NIL;
  Pt.Size:=0;
  if MainMap[Y,X].Distance < 0 then Exit;
  if Length(Pt.Path)<MainMap[Y,X].Distance+1 then begin
    SetLength(Pt.Path,MainMap[Y,X].Distance+256); //    .
  end;
  Pt.Size:=MainMap[Y,X].Distance+1;
  //           
  //      .
  while MainMap[Y,X].Distance > 0 do
    begin
    Pt.Path[MainMap[Y,X].Distance]:=Point(X,Y);
    Direction:=MainMap[Y,X].Direction;
    X:=X-DirToDX(Direction);
    Y:=Y-DirToDY(Direction);
    end;
  Pt.Path[0]:=Point(X,Y);  //    .
end;

procedure  FindPath(
  var pt:TPathGroup;
  StartX,StartY,StopX,StopY : Integer;
  GetCostFunc : TGetCostFunc);
begin
  PathFind.GetCostFunc:=GetCostFunc;
  //             .
  //    .
  FillPathMap(StartX,StartY,StopX,StopY);
  FindPathOnMap(pt,StopX,StopY);
end;

end.
